/*!
 * © 2014, 2022 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals Character: true, alert, Rules, Genre, options, saveAs */
Character = {};
var character;

Character.allstats = [ "str", "dex", "agi", "con", "app", "int", "wit", "will" ];
Character.longstats = [ "Strength", "Dexterity", "Agility", "Constitution",
        "Appearance", "Intellect", "Wit", "Will" ];

Character.returnFalse = function(f) {
    return function() {
        try {
            f();
        } catch (e) {
            if (e.stack) {
                alert(e + "\n\nStack Trace\n-----------------\n" + e.stack);
            } else {
                alert(e);
            }
        }
        return false;
    };
};

Character.incrementVersion = function(cid) {
    character[cid].characterVersion += 1;
    character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
    Character.populateJSON(cid,true);
};

Character.updateDays = function(cid) {
    character[cid].birthday = character[cid].genre.dayMonthYearAsText(character[cid].birthtime, character[cid].birthmonth);
    character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
};

Character.templates = {
    interest : {
        skill : "Numeracy",
        ip : 1,
        teacher : false
    }
};

Character.updateNode = function(node,i) {
    if (node.nodeType === Node.ELEMENT_NODE) {
        node.id = node.id.replace("@@",i.toString());
        if (node.name !== undefined) {
            node.name = node.name.replace("@@",i.toString());
        }
        for(var j = 0; j < node.childNodes.length; j++) {
            Character.updateNode(node.childNodes[j],i);
        }
    }
};

Character.updateForm = function(node,i) {
    if (node.nodeType === Node.ELEMENT_NODE) {
        node.id = node.id.replace("-interest0-","-interest"+i+"-");
        if (node.name !== undefined) {
            node.name = node.name.replace("-interest0-","-interest"+i+"-");
        }
        for(var j = 0; j < node.childNodes.length; j++) {
            Character.updateForm(node.childNodes[j],i);
        }
    }
};

Character.setIdText = function(cid, id, text, tooltip) {
    var element = document.getElementById("c"+cid+"-"+id);
    var newText = document.createTextNode(text);
    if (tooltip) {
        element.title = tooltip;
    }
    if (element.hasChildNodes()) {
        element.replaceChild(newText, element.firstChild);
    } else {
        element.appendChild(newText);
    }
};
Character.setIdClass = function(cid, id, cls) {
    var element = document.getElementById("c"+cid+"-"+id);
    element.className = cls;
};
Character.setTextArea = function(cid, field, text) {
    var element = document.getElementById("c"+cid+"-"+field);
    element.value = text;
    element.defaultValue = text;
};

Character.setOption = function(options, choice) {
    var found = false;
    for (var j = 0; j < options.length; j++) {
        if (choice.toString() === options[j].text) {
            options.selectedIndex = j;
            found = true;
        }
    }
    if (!found) {
        var newOption = new Option(choice.toString(), choice.toString(), true,
                true);
        options.add(newOption);
    }
};

Character.setChecked = function(button, choice) {
    button.checked = choice;
    button.defaultChecked = choice;
};

Character.deleteElement = function(cid, idOrElement) {
    var element = ((typeof idOrElement === "string") ? document.getElementById("c"+cid+"-"+idOrElement) : idOrElement);
    if (element) {
        element.parentNode.removeChild(element);
    }
};

Character.durationLength = function(dur) {
    if (isNaN(dur)) {
        throw new Error("Duration is not a number");
    } else if (dur === 0) {
        return "0 months";
    } else if (dur === 1) {
        return "1 month";
    } else if (dur < 12) {
        return dur + " months";
    }
    var years = Math.floor(dur / 12);
    dur = dur % 12;

    if (years === 1) {
        if (dur === 0) {
            return "1 year";
        }
        return "1 year, " + Character.durationLength(dur);
    } else {
        if (dur === 0) {
            return years + " years";
        }
        return years + " years, " + Character.durationLength(dur);
    }
};

Character.getFluencyPoints = function(cid,p) {
    var endFluency = (character[cid].periods[p].endmonth < 179) ? character[cid].periods[p].endmonth
            : 179;
    var fluencyMonths = endFluency - character[cid].periods[p].startmonth + 1;
    if (fluencyMonths < 0) {
        return 0;
    }
    var points = (character[cid].periods[p].startmonth === 60) ? 1000 : 0;
    points += fluencyMonths * 8;
    for (var anniversary = 72; anniversary <= 180; anniversary += 12) {
        if ((character[cid].periods[p].startmonth <= anniversary) && (character[cid].periods[p].endmonth >= anniversary)) {
            points += 4;
        }
    }
    return points;
};

Character.updateFluencyPoints = function(cid,p) {
    var points = Character.getFluencyPoints(cid,p);
    if (points > 0) {
        Character.setIdText(cid,"period"+p+"-fluencypoints",points+" points");
    }
};

Character.deleteAllChildren = function(node) {
    while (node.firstChild) {
        node.removeChild(node.firstChild);
    }
};

Character.addInterestRow = function(cid, p,table,index,ineditmode) {
    var template = document.getElementById("c"+cid+"-period@@-interest0");
    var row = template.cloneNode(true);
    var next = index;
    Character.updateNode(row,p);
    Character.updateForm(row,next);
    row.id = "c"+cid+"-period"+p+"-interest"+next;
    table.appendChild(row);
    Character.addSkillRow(cid,p,next,ineditmode);
    Character.populateJSON(cid,ineditmode);
};

Character.addOtherSkill = function(cid,name,stats,ineditmode) {
    if (name !== "") {
        Rules.addSkill(name,stats);

        if (character[cid].newSkills.every(function(x) { return x[name] === undefined; })) {
            var newSkills = {};
            newSkills[name] = stats;
            character[cid].newSkills.push(newSkills);
        }

        // Finally add a button for it - if the table for the buttons exists
        if (ineditmode) {
            const DOM_div = document.getElementById("c"+cid+"-skills-category-"+encodeURI("Added skills"));
            if (DOM_div) {
                const DOM_h4 = DOM_div.firstChild;
                var DOM_table, DOM_tbody, DOM_tr, DOM_td;
                if (DOM_h4) {
                    DOM_table = DOM_h4.nextSibling;
                    DOM_tbody = DOM_table.firstChild;
                    if (!DOM_tbody.lastChild || DOM_tbody.lastChild.lastChild.lastChild) {
                        // Empty space on last row
                        DOM_tr = document.createElement("tr");
                        DOM_td = document.createElement("td");
                        DOM_td.appendChild(Character.createButton(name,Character.addInterest(cid,name,true)));
                        DOM_tr.appendChild (DOM_td);
                        DOM_td = document.createElement("td");
                        DOM_tr.appendChild(DOM_td);
                        DOM_tbody.appendChild(DOM_tr);
                    } else {
                        DOM_td = DOM_tbody.lastChild.lastChild;
                        DOM_td.appendChild(Character.createButton(name,Character.addInterest(cid,name,true)));
                    }
                }
            }
        }
    }

    // And give the character the skill if they're on a period
    Character.addInterest(cid,name,ineditmode)();
};

Character.addSkillRow = function(cid,p,interest,ineditmode) {
    Character.updateSkillRow(cid,p,interest,ineditmode);
    if (ineditmode) {
        var interestup = document.getElementById("c"+cid+"-period"+p+"-interest"+interest+"-ipup");
        var interestdown = document.getElementById("c"+cid+"-period"+p+"-interest"+interest+"-ipdown");
        var interestteacher = document.getElementById("c"+cid+"-period"+p+"-interest"+interest+"-teacher");
        var interestdelete = document.getElementById("c"+cid+"-period"+p+"-interest"+interest+"-delete");
        interestup.onclick = Character.updatePoints(cid,p,interest,1);
        interestdown.onclick = Character.updatePoints(cid,p,interest,-1);
        interestteacher.onchange = Character.updateTeacher(cid,p,interest,interestteacher);
        interestdelete.onclick = Character.deleteInterest(cid,p,interest);
    }
};

Character.updateSkillRow = function(cid,p,interest,ineditmode) {
    var interestteacher = document.getElementById("c"+cid+"-period"+p+"-interest"+interest+"-teacher");
    Character.setIdText(cid,"period"+p+"-interest"+interest+"-skill",character[cid].periods[p].interests[interest].skill);
    Character.setIdText(cid,"period"+p+"-interest"+interest+"-points",character[cid].periods[p].interests[interest].ip);
    if (ineditmode) {
        Character.setChecked(interestteacher,character[cid].periods[p].interests[interest].teacher);
    } else {
        Character.setIdClass(cid,"period"+p+"-interest"+interest+"-plural",character[cid].periods[p].interests[interest].ip == 1 ? "hidden" : "");
        if (character[cid].periods[p].interests[interest].teacher) {
            Character.setIdText(cid,"period"+p+"-interest"+interest+"-teacher","Teacher Bonus");
        }
    }
};

Character.addInterest = function(cid,skillname,ip,ineditmode) {
    return Character.returnFalse(function() {
        // Work out which tab is active
        var tab;
        var tabbersheet = document.getElementById("c"+cid+"-sheet");
        if (tabbersheet && tabbersheet.tabber) {
            tab = tabbersheet.tabber.getActiveTab();
        }
        if (tab) {
            ip = ip || 1;
            var idbits = tab.id.split("-"); // Note we assume an ordering of cid, period, interest as set up by addInterestRow
            if (idbits[1].slice(0,6) === "period") {
                var p = Number(idbits[1].slice(6));
                var table = document.getElementById("c"+cid+"-period"+p+"-details").firstChild.nextSibling;
                if (character[cid].periods[p].iptotal+ip <= character[cid].interestPoints) {
                    Character.addInterestKnowingPeriod(cid,p,table,skillname,ip,ineditmode);
                    Character.incrementVersion(cid);
                    Character.updateStatus(cid,p);
                }
            }
        }
    });
};

Character.addInterestKnowingPeriod = function(cid,p,table,skillname,ip,ineditmode) {
    var adding = true;
    for(var i=0; i<character[cid].periods[p].interests.length; i++) {
        if (character[cid].periods[p].interests[i].skill === skillname) {
            character[cid].periods[p].iptotal += ip;
            character[cid].periods[p].interests[i].ip += ip;
            Character.updateSkillRow(cid,p,i);
            adding = false;
        }
    }
    if (adding) {
        var next = character[cid].periods[p].interests.length;
        character[cid].periods[p].interests.push(JSON.parse(JSON.stringify(Character.templates.interest)));
        character[cid].periods[p].interests[next].skill = skillname;
        character[cid].periods[p].interests[next].ip = ip;
        character[cid].periods[p].iptotal += ip;
        Character.addInterestRow(cid,p,table,next,ineditmode);
    }
    if (ineditmode) Character.updateSkillRowButtons(cid,p);
    Character.populateSkills(cid,ineditmode);
    Character.populateJSON(cid,ineditmode);
};

Character.updateTabs = function(cid) {
    var tabbersheet = document.getElementById("c"+cid+"-sheet");
    if (tabbersheet.tabber) {
        tabbersheet.tabber.reinit(tabbersheet);
    }
};

Character.updateArmour = function(cid,armour) {
    return Character.returnFalse(function() {
        character[cid].armour = armour;
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });
};

Character.updateShield = function(cid,shield) {
    return Character.returnFalse(function() {
        character[cid].shield = shield;
        if (character[cid].shield.locations === 0) {
            character[cid].shieldLocations = [];
        } else if (character[cid].shield.locations === 1) {
            character[cid].shieldLocations = [ "shieldArm", "chest"];
        } else if (character[cid].shield.locations === 2) {
            character[cid].shieldLocations = [ "shieldArm", "chest","head"];
        } else if (character[cid].shield.locations === 3) {
            character[cid].shieldLocations = [ "shieldArm", "chest","head","abdomen"];
        }
        Character.populateShieldLocations(cid);
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });
};

Character.updateShieldLocations = function(cid) {
    return Character.returnFalse(function() {
        var locs;
        locs = ["head", "chest", "abdomen"];
        character[cid].shieldLocations = ["shieldArm"];
        for(var i=0; i<locs.length; i++) {
            if (document.getElementById("c"+cid+"-coverage"+locs[i]).checked) {
                character[cid].shieldLocations.push(locs[i]);
            }
        }
        Character.incrementVersion(cid);
    });
};

Character.populateShieldLocations = function(cid) {
    var containsShieldLocation = function(location) {
        return function(x) { return x === location; };
    };
    document.getElementById("c"+cid+"-trcoverageinfo").firstChild.data = character[cid].shield.text;
    var head =  document.getElementById("c"+cid+"-coveragehead");
    var chest = document.getElementById("c"+cid+"-coveragechest");
    var abdomen = document.getElementById("c"+cid+"-coverageabdomen");
    if (character[cid].shield.locations === 0) {
        head.type = "checkbox";
        chest.type = "checkbox";
        abdomen.type = "checkbox";
        head.checked = false;
        chest.checked = false;
        abdomen.checked = false;
        head.disabled = true;
        chest.disabled = true;
        abdomen.disabled = true;
    } else if (character[cid].shield.locations === 1) {
        head.type = "radio";
        chest.type = "radio";
        abdomen.type = "radio";
        head.checked = character[cid].shieldLocations.some(containsShieldLocation("head"));
        chest.checked = character[cid].shieldLocations.some(containsShieldLocation("chest"));
        abdomen.checked = character[cid].shieldLocations.some(containsShieldLocation("abdomen"));
        head.disabled = false;
        chest.disabled = false;
        abdomen.disabled = false;
        head.onchange = Character.updateShieldLocations(cid);
        chest.onchange = Character.updateShieldLocations(cid);
        head.onchange = Character.updateShieldLocations(cid);
    } else if (character[cid].shield.locations === 2) {
        head.type = "radio";
        chest.type = "checkbox";
        abdomen.type = "radio";
        head.checked = character[cid].shieldLocations.some(containsShieldLocation("head"));
        chest.checked = true;
        abdomen.checked = character[cid].shieldLocations.some(containsShieldLocation("abdomen"));
        head.disabled = false;
        chest.disabled = true;
        abdomen.disabled = false;
        head.onchange = Character.updateShieldLocations(cid);
        abdomen.onchange = Character.updateShieldLocations(cid);
    } else if (character[cid].shield.locations === 3) {
        head.type = "checkbox";
        chest.type = "checkbox";
        abdomen.type = "checkbox";
        head.checked = true;
        chest.checked = true;
        abdomen.checked = true;
        head.disabled = true;
        chest.disabled = true;
        abdomen.disabled = true;
    }
};

Character.updateFavouredWeapons = function(cid) {
    return Character.returnFalse(function() {
        character[cid].favouredWeapons = [];
        for(var i=0; i<Rules.weapons.length; i++) {
            var checkbox = document.getElementById("c"+cid+"-weapon-"+Rules.weapons[i].name);
            if (checkbox && checkbox.checked) {
                character[cid].favouredWeapons.push(Rules.weapons[i]);
            }
        }
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });
};

Character.populateEquipment = function(cid,ineditmode) {
    var containsWeapon = function(weapon) { return function(x) { return x.name === weapon.name; }; };
    var containsGenre = function(x) { return x === character[cid].genre.toString(); };
    var radioequip = function(radioname,array,setting,ineditmode,update) {
        var b = 0;
        var button,span;
        for(var i=0; i<array.length; i++) {
            if (array[i].genre.some(containsGenre)) {
                span = document.getElementById("c"+cid+"-td"+radioname+b);
                button = document.getElementById("c"+cid+"-"+radioname+b);
                span.className = "";
                button.nextSibling.data = array[i].name;
                button.checked = (array[i].name === setting.name);
                if (ineditmode)
                    button.onchange = update(cid,array[i]);
                else
                    button.disabled = true;
                b++;
            }
        }
        // Clear out the other buttons, if any
        for(;;b++) {
            span = document.getElementById("c"+cid+"-td"+radioname+b);
            if (!span)  { break; }
            button = document.getElementById("c"+cid+"-"+radioname+b);
            span.className = "hidden";
            if (button.checked) {
                button.checked = false;
                document.getElementById("c"+cid+"-"+radioname+"0").checked = true;
            }
        }

    };

    radioequip("armour",Rules.armour,character[cid].armour,ineditmode,Character.updateArmour);
    radioequip("shield",Rules.shields,character[cid].shield,ineditmode,Character.updateShield);

    // shield locations
    Character.populateShieldLocations(cid);

    // Weapons are a set of checkboxs, (6 columns)
    var columns = 6;
    var weapons = document.getElementById("c"+cid+"-weapons");
    var newweapons = document.createDocumentFragment();
    var newrow = document.createElement("tr");
    newrow.id = "c"+cid+"-weapons";
    var column = 0;
    for(var i=0; i<Rules.weapons.length; i++) {
        if (Rules.weapons[i].genre.some(containsGenre)) {
            var newcheckbox = document.createElement("input");
            newcheckbox.type = "checkbox";
            newcheckbox.id = "c"+cid+"-weapon-"+Rules.weapons[i].name;
            newcheckbox.checked = character[cid].favouredWeapons.some(containsWeapon(Rules.weapons[i]));
            if (ineditmode)
                newcheckbox.onchange = Character.updateFavouredWeapons(cid);
            else
                newcheckbox.disabled = true;
            var newcell = document.createElement("td");
            newcell.appendChild(newcheckbox);
            newcell.appendChild(document.createTextNode(Rules.weapons[i].name));
            newrow.appendChild(newcell);
            column++;
            if (column === columns) {
                newweapons.appendChild(newrow);
                newrow = document.createElement("tr");
                column = 0;
            }
        }
    }
    if (column > 0) {
        newweapons.appendChild(newrow);
    }
    while(weapons.nextSibling) {
        weapons.parentNode.removeChild(weapons.nextSibling);
    }
    weapons.parentNode.replaceChild(newweapons,weapons);
};

Character.populatePeriodMonths = function(cid,p) {
    Character.setIdText(cid,"period"+p+"-enddatelength"," A period of "+Character.durationLength(character[cid].periods[p].endmonth - character[cid].periods[p].startmonth  + 1));
    if (character[cid].periods[p].startmonth === character[cid].periods[p].endmonth) {
        Character.setIdText(cid,"period"+p+"-daterange","The month of "+character[cid].periods[p].startmonthtext);
    } else if (character[cid].periods[p].startmonth < character[cid].periods[p].endmonth) {
        Character.setIdText(cid,"period"+p+"-daterange",character[cid].periods[p].startmonthtext+" to "+character[cid].periods[p].endmonthtext);
    } else {
        Character.setIdText(cid,"period"+p+"-daterange","Future interests");
    }
};

Character.populateStats = function(cid,ineditmode) {
    var i;
    Character.setTextArea(cid,"nametextarea",character[cid].name);
    Character.setTextArea(cid,"playernametextarea",character[cid].playerName);
    if (ineditmode) {
        Character.setOption(document.getElementById("c"+cid+"-genre"),character[cid].genre.toString());
        Character.setOption(document.getElementById("c"+cid+"-gender"),character[cid].gender.toString());
    }
    for(i=0; i<Character.allstats.length; i++) {
        Character.setIdText(cid,"stats-"+Character.allstats[i],character[cid][Character.allstats[i]]);
    }

    Character.setIdText(cid,"stats-morale",character[cid].morale);
    Character.setIdText(cid,"stats-ip",character[cid].interestPoints);

	if (ineditmode)
        Character.setTextArea(cid,"characterdescriptiontextarea",character[cid].description);
    else
        Character.setIdText(cid,"characterdescriptiontextarea",character[cid].description);

    Character.setIdText(cid,"stats-birthday",character[cid].birthday);
    Character.setIdText(cid,"stats-gameday",character[cid].gameday);

    if (ineditmode) {
        document.getElementById("c"+cid+"-reroll").disabled = !(character[cid].statsswapable);
        document.getElementById("c"+cid+"-statsedit").disabled = character[cid].statslock || !(character[cid].statsswapable);
        document.getElementById("c"+cid+"-statslock").disabled = character[cid].statslock;
        document.getElementById("c"+cid+"-statsunlock").disabled = !(character[cid].statslock);

        document.getElementById("c"+cid+"-birthdayminus1year").disabled = character[cid].statslock;
        document.getElementById("c"+cid+"-birthdayminus1month").disabled = character[cid].statslock;
        document.getElementById("c"+cid+"-birthdayminus1day").disabled = character[cid].statslock;
        document.getElementById("c"+cid+"-birthdayplus1day").disabled = character[cid].statslock;
        document.getElementById("c"+cid+"-birthdayplus1month").disabled = character[cid].statslock;
        document.getElementById("c"+cid+"-birthdayplus1year").disabled = character[cid].statslock;

        document.getElementById("c"+cid+"-gamedatesync").checked = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedatesync").defaultChecked = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayminus1year").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayminus1month").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayminus1day").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayplus1day").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayplus1month").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayplus1year").disabled = character[cid].syncWithGlobalDate;

        if (!character[cid].statsswapable) {
            for(i=0; i<Character.allstats.length; i++) {
                Character.deleteElement(cid,Character.allstats[i]+"swap");
            }
        }
    }
};

Character.populatePeriods = function(cid,ineditmode) {
    // Delete the old periods
    var i = 0;
    var period = document.getElementById("c"+cid+"-period"+i);
    while(period) {
        Character.deleteElement(cid,period);
        i++;
        period = document.getElementById("c"+cid+"-period"+i);
    }

    // And build new ones
    var periods = document.getElementById("c"+cid+"-periods");
    var template = document.getElementById("c"+cid+"-period@@");

    for(i = 0; i < character[cid].periods.length; i++) {
        var newNode = template.cloneNode(true);
        newNode.className = "tabbertab";
        newNode.title = character[cid].periods[i].title;
        Character.updateNode(newNode,i);
        periods.parentNode.insertBefore(newNode,periods);

        // Header fields
        if (ineditmode) {
            Character.setTextArea(cid,"period"+i+"-titletextarea",character[cid].periods[i].title);
            Character.setTextArea(cid,"period"+i+"-descriptiontextarea",character[cid].periods[i].description);
        } else {
            Character.setIdText(cid,"period"+i+"-descriptiontextarea",character[cid].periods[i].description);
        }

        // dates
        Character.setIdText(cid,"period"+i+"-startdate","Age: "+Character.durationLength(character[cid].periods[i].startmonth),character[cid].genre.dateMonthAsText(character[cid].periods[i].startmonth+character[cid].birthmonth));
        Character.setIdText(cid,"period"+i+"-enddate","Age: "+Character.durationLength(character[cid].periods[i].endmonth),character[cid].genre.dateMonthAsText(character[cid].periods[i].endmonth+character[cid].birthmonth));
        Character.setIdText(cid,"period"+i+"-enddatelength"," A period of "+Character.durationLength(character[cid].periods[i].endmonth - character[cid].periods[i].startmonth  + 1));
        if (ineditmode) {
            document.getElementById("c"+cid+"-period"+i+"-titletextarea").onchange = Character.populatePeriodTitle(cid,i,i+3);
            document.getElementById("c"+cid+"-period"+i+"-descriptiontextarea").onchange = Character.populatePeriodDescription(cid,i);
            document.getElementById("c"+cid+"-period"+i+"-enddateminus1year").onclick = Character.updateEndDate(cid,i,-12);
            document.getElementById("c"+cid+"-period"+i+"-enddateminus1month").onclick = Character.updateEndDate(cid,i,-1);
            document.getElementById("c"+cid+"-period"+i+"-enddateplus1month").onclick = Character.updateEndDate(cid,i,1);
            document.getElementById("c"+cid+"-period"+i+"-enddateplus1year").onclick = Character.updateEndDate(cid,i,12);
        }

        Character.populatePeriodMonths(cid,i);

        // childhood fluency
        if (character[cid].periods[i].startmonth > 180) {
            Character.deleteElement(cid,document.getElementById("c"+cid+"-period"+i+"-fluencyheader"));
            Character.deleteElement(cid,document.getElementById("c"+cid+"-period"+i+"-fluencycontent"));
        } else {
            Character.updateFluencyPoints(cid,i);
            if (ineditmode) {
                var fluencyChoice = document.getElementById("c"+cid+"-period"+i+"-childhoodfluencylanguage");
                Character.setOption(fluencyChoice,character[cid].periods[i].childhoodLanguage);
                fluencyChoice.onchange = Character.updateFluency(cid,i,fluencyChoice);
            } else {
                Character.setIdText(cid,"period"+i+"-childhoodfluencylanguage",character[cid].periods[i].childhoodLanguage);
            }
        }

        // Interests
        var table = document.getElementById("c"+cid+"-period"+i+"-details").firstChild.nextSibling;
        for(var interest=0; interest < character[cid].periods[i].interests.length; interest++) {
            Character.addInterestRow(cid,i,table,interest,ineditmode);
        }
        if (ineditmode) {
            Character.updateSkillRowButtons(cid,i);
            Character.setIdClass(cid,"period"+i+"-copy-previous",(i==0 ? "hidden" : ""));
            document.getElementById("c"+cid+"-period"+i+"-copy-previous").onclick = Character.copyPreviousPeriodSkills(cid,i);
        }

        Character.updateStatus(cid,i);
    }

    Character.updateTabs(cid);
};

Character.populateSkills = function(cid,ineditmode) {
    var i,j;
    var endingMonthText = character[cid].genre.dateMonthAsText(character[cid].periods[character[cid].periods.length-1].endmonth+character[cid].birthmonth);

    var points = {};
    var addPoints = function(points,name,dp) {
        if (points[name]) {
            points[name] += dp;
        } else {
            points[name] = dp;
        }
    };
    for(i=0; i<character[cid].periods.length; i++) {
        // Childhood fluency
        if (Character.getFluencyPoints(cid,i) > 0) {
            addPoints(points,"Fluency("+character[cid].periods[i].childhoodLanguage+")",Character.getFluencyPoints(cid,i));
        }
        var duration = character[cid].periods[i].endmonth - character[cid].periods[i].startmonth + 1;
        for(j=0; j<character[cid].periods[i].interests.length; j++) {
            var dp = character[cid].periods[i].interests[j].ip * duration;
            if (character[cid].periods[i].interests[j].teacher) {
                dp += (Math.round(character[cid].periods[i].interests[j].ip / 3 - 0.3)) * duration;
            }
            addPoints(points,character[cid].periods[i].interests[j].skill,dp);
        }
    }
    // convert to array and sort
    var arr = [];
    for(var s in points) {
        if (points.hasOwnProperty(s)) {
            arr.push({skill:s, dp: points[s]});
        }
    }
    // Compute the skill levels
    for(i=0; i<arr.length; i++) {
        arr[i].level = Rules.skilllevel(arr[i].skill,character[cid],arr[i].dp);
    }


    arr.sort(function(x,y) {
        return y.level - x.level;
    });
    var DOM_table = document.createElement("table");
    var DOM_tbody = document.createElement("tbody");
    var DOM_tr = document.createElement("tr");
    var DOM_th = document.createElement("th");
    DOM_th.colspan = 3;
    var DOM_p = document.createElement("p");
    DOM_p.appendChild(document.createTextNode("As of the month of "+endingMonthText));
    DOM_th.appendChild(DOM_p);
    DOM_tr.appendChild(DOM_th);
    DOM_tbody.appendChild(DOM_tr);

    DOM_tr = document.createElement("tr");
    DOM_th = document.createElement("th");
    DOM_th.appendChild(document.createTextNode("Skill Name"));
    DOM_tr.appendChild(DOM_th);
    DOM_th = document.createElement("th");
    DOM_th.appendChild(document.createTextNode("Development Points"));
    DOM_tr.appendChild(DOM_th);
    DOM_th = document.createElement("th");
    DOM_th.appendChild(document.createTextNode("Level"));
    DOM_tr.appendChild(DOM_th);
    DOM_tbody.appendChild(DOM_tr);

    for(i=0; i<arr.length; i++) {
        DOM_tr = document.createElement("tr");
        var DOM_td = document.createElement("td");
        DOM_td.appendChild(document.createTextNode(arr[i].skill));
        DOM_tr.appendChild(DOM_td);
        DOM_td = document.createElement("td");
        DOM_td.style = "text-align: right";
        DOM_td.appendChild(document.createTextNode(arr[i].dp));
        DOM_tr.appendChild(DOM_td);
        DOM_td = document.createElement("td");
        DOM_td.style = "text-align: right";
        DOM_td.appendChild(document.createTextNode(arr[i].level));
        DOM_tr.appendChild(DOM_td);
        DOM_tbody.appendChild(DOM_tr);
    }
    DOM_table.appendChild(DOM_tbody);

    var skills = document.getElementById("c"+cid+"-skillsheet");
    Character.deleteAllChildren(skills);
    skills.appendChild(DOM_table);

    // And now the JSON needs updating
    character[cid].sheet = { endmonth: character[cid].periods[character[cid].periods.length-1].endmonth,
                        endmonthtext: endingMonthText,
                        skills: arr
                      };
    Character.populateJSON(cid,ineditmode);
};

Character.populateJSON = function(cid,ineditmode) {
    if (ineditmode) {
        Character.setTextArea(cid,"JSONtextarea",JSON.stringify(character[cid],undefined,2));
    }
};

Character.validateStat = function(name,stat) {
    if (stat === undefined) { throw name+" is blank, character has not been updated."; }
    var v = Number(String(stat).trim());
    if (isNaN(v)) { throw name+" is not a number (it's '"+stat+"'). character has not been updated."; }
    if (v < 0)    { throw name+" is less than zero, character has not been updated."; }
    if (v > 20)   { throw name+" is more than twenty, character has not been updated."; }
    return v;
};

Character.createButton = function(name,onclickfunction) {
    var DOM_button = document.createElement("button");
    var text = name.replace("/","/\u200b").replace("(","\u200b(");
    DOM_button.appendChild(document.createTextNode(text));
    DOM_button.value = name;
    DOM_button.style.width = "100%";
    DOM_button.onclick = onclickfunction;
    return DOM_button;
};

Character.createTextArea = function(cid, id, rows, cols) {
    var DOM_textarea = document.createElement("textarea");
    DOM_textarea.id= "c"+cid+"-"+id;
    DOM_textarea.rows = rows;
    DOM_textarea.cols = cols;
    if (rows === 1) {
        DOM_textarea.className = "onerow";
    }
    DOM_textarea.appendChild(document.createTextNode(""));
    return DOM_textarea;
};

Character.updateStatus = function(cid,p) {
    var status = document.getElementById("c"+cid+"-period" + p + "-status");
    var text;
    var valid = false;
    var DOM_b;
    Character.deleteAllChildren(status);
    if (character[cid].periods[p].endmonth === character[cid].periods[p].startmonth-1) {
        status.setAttribute("class", "warning");
        text = "This period is of zero length and will not add any development points to this character's skills.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].endmonth < character[cid].periods[p].startmonth) {
        status.setAttribute("class", "error");
        text = "This period is of NEGATIVE length and will REMOVE development points from this character's skills.";
        DOM_b = document.createElement("b");
        DOM_b.appendChild(document.createTextNode(text));
        status.appendChild(DOM_b);
    } else if (p === character[cid].periods[p].length-1 && character[cid].periods[p].endMonth !== character[cid].gamemonth-character[cid].birthmonth) {
        status.setAttribute("class", "warning");
        text = "The end month of this last period (" + character[cid].periods[p].endMonthText + ") does not match the game month (" + character[cid].gamemonthtext + "). When the game month is updated, this will reset this end month to the game month.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].iptotal === 0) {
        status.setAttribute("class", "warning");
        text = "None of the " + character[cid].interestPoints + " interest points available to this character have been used.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].iptotal < character[cid].interestPoints) {
        status.setAttribute("class", "warning");
        text = "Only " + character[cid].periods[p].iptotal + " of the " + character[cid].interestPoints + " interest points available to this character have been used.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].iptotal === character[cid].interestPoints) {
        status.setAttribute("class", "");
        text = "All the " + character[cid].interestPoints + " interest points available to this character have been used.";
        valid = true;
        status.appendChild(document.createTextNode(text));
    } else {
        status.setAttribute("class", "error");
        text = "The " + character[cid].periods[p].iptotal + " interest points used by this character are more than the " + character[cid].interestPoints + " available to this character[cid].";
        DOM_b = document.createElement("b");
        DOM_b.appendChild(document.createTextNode(text));
        status.appendChild(DOM_b);
    }
    character[cid].periods[p].statustext = text;
    character[cid].periods[p].valid = valid;
};

Character.populateAll = function(cid,ineditmode) {
    Character.populateStats(cid,ineditmode);
    Character.populateEquipment(cid,ineditmode);
    Character.populatePeriods(cid,ineditmode);
    Character.populateSkills(cid,ineditmode);
};


Character.initPage = function(cid) {
    // Initialise the character, either with data from an embedded JSON or a completely new one
    character = character || [];
    // Set up all the non-Period buttons (and work out if we are editting)
    const ineditmode = Character.setupButtons(cid);

    if (document.getElementById("c"+cid+"-character-data") && document.getElementById("c"+cid+"-character-data").text && document.getElementById("c"+cid+"-character-data").text.length > 10) {
        Character.updateWithNewJSON(cid, document.getElementById("c"+cid+"-character-data").text, false, ineditmode);
    } else {
        character[cid] = Rules.newCharacter();
        character[cid].schemaVersion = 13;
        character[cid].characterVersion = 0;
        character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
        character[cid].rulesVersion = "2014a";
        character[cid].genre = new Genre(options.default_genre);
        character[cid].name = "Nobody Noname";
        character[cid].gender = '';
        character[cid].playerName = document.getElementById("c"+cid+"-playernametextarea").textContent;
        character[cid].description = "An outstanding character whose deeds are described herein and whose heroic future is still to come.";
        character[cid].statslock = false;
        character[cid].statsswapable = true;
        character[cid].morale = Rules.startingMorale(character[cid]);
        character[cid].woundPoints = Rules.woundPoints(character[cid]);
        character[cid].interestPoints = Rules.interestPoints(character[cid]);
        character[cid].birthtime = Date.now() - 32 * 365.25 * 24 * 60 * 60 * 1000;
        character[cid].birthmonth = character[cid].genre.asMonth(character[cid].birthtime);
        character[cid].birthmonthtext = character[cid].genre.dateMonthAsText(character[cid].birthmonth);
        character[cid].birthday = character[cid].genre.dayMonthYearAsText(character[cid].birthtime, character[cid].birthmonth);
        character[cid].gametime = options.gametime[character[cid].genre.name];
        character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
        character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
        character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime, character[cid].gamemonth);
        character[cid].periods = [ {
            title : "Childhood",
            description : "The early years discovering the world and your place in it",
            startmonth : 60,
            startmonthtext : character[cid].genre.dateMonthAsText(character[cid].birthmonth + 60),
            endmonth : character[cid].gamemonth - character[cid].birthmonth,
            endmonthtext : character[cid].gamemonthtext,
            childhoodLanguage : character[cid].genre.getDefaultLanguage(),
            interests : [],
            iptotal : 0,
            statustext : "None of the " + character[cid].interestPoints + " interest points available to this character have been used.",
            valid : false
        } ];
        character[cid].newSkills = [];
        character[cid].armour = Rules.armour[0];
        character[cid].shield = Rules.shields[0];
        character[cid].shieldLocations = [];
        character[cid].favouredWeapons = [];
        character[cid].syncWithGlobalDate = true;
    }
    // And the save buttons (if present)
    const savejson = document.getElementById("c"+cid+"-JSONsavefile");
    if (savejson) {
        savejson.onclick = Character.returnFalse(function() {
                    Character.incrementVersion(cid);
                    var blob = new Blob(
                            [ JSON.stringify(character[cid], undefined, 2) ], {
                                type : "text/plain;charset=utf-8"
                            });
                    saveAs(blob, character[cid].name + ".txt");
                });
    }

    const savepdf = document.getElementById("c"+cid+"-savePDF");
    if (savepdf) {
        savepdf.onclick = Character.returnFalse(function() {
                    var blob = Character.makeOneCharacterPDF(character[cid]).output('blob');
                    saveAs(blob, character[cid].name + ".pdf");
                });
     }

     return ineditmode;
};

Character.resetOtherSkillsList = function(cid,ineditmode) {
    if (ineditmode) {
        const DOM_div = document.getElementById("c"+cid+"-skills-category-"+encodeURI("Added skills"));
        if (DOM_div) {
            const DOM_h4 = DOM_div.firstChild;
            if (DOM_h4) {
                var DOM_table = DOM_h4.nextSibling;
                var DOM_tbody = DOM_table.firstChild;
                Character.deleteAllChildren(DOM_tbody);
            }
        }
    }
};

Character.updateWithNewJSON = function(cid,json,populate,ineditmode) {
    try {
        var i;
        var newCharacter = JSON.parse(json);
        for(i=0; i<Character.allstats.length; i++) {
            newCharacter[Character.allstats[i]] = Character.validateStat(Character.allstats[i],newCharacter[Character.allstats[i]]);
        }
        character[cid] = newCharacter;

        if (!character[cid].genre) {
            // Need to guess! Base off of the childhood language in the first period
            if (character[cid].periods[0].childhoodLanguage === "Athionic") {
                character[cid].genre = new Genre("Linrodeth");
            } else if (character[cid].periods[0].childhoodLanguage === "Latin") {
                character[cid].genre = new Genre("Luna Romana");
            } else if (character[cid].periods[0].childhoodLanguage === "English") {
                character[cid].genre = new Genre("Contemporary");
            } else if (character[cid].periods[0].childhoodLanguage === "Tsolyáni") {
                character[cid].genre = new Genre("Tekumel");
            } else if (character[cid].periods[0].childhoodLanguage === "Anglic") {
                character[cid].genre = new Genre("Traveller");
            } else {
                character[cid].genre = new Genre("Linrodeth");
            }
        } else {
            character[cid].genre = new Genre(character[cid].genre);
        }

        character[cid].morale = Rules.startingMorale(character[cid]);
        character[cid].woundPoints = Rules.woundPoints(character[cid]);
        character[cid].interestPoints = Rules.interestPoints(character[cid]);
        if (!character[cid].schemaVersion) {
            character[cid].schemaVersion = 1;
            character[cid].characterVersion = 1e9; // Large offset to make it obvious
            character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
            character[cid].rulesVersion = "2014a";
            character[cid].description = "An outstanding character whose deeds are described herein.";
            character[cid].gametime = character[cid].birthtime + character[cid].periods[character[cid].periods.length-1].endmonth*30.4375*24*60*60*1000; // Approx
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].birthmonth);
            character[cid].gamemonth = character[cid].periods[character[cid].periods.length-1].endmonth + character[cid].birthmonth;
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
            character[cid].periods[character[cid].periods.length-1].endMonth = character[cid].gamemonth;
            character[cid].periods[character[cid].periods.length-1].endMonthText = character[cid].gamemonthtext;
            character[cid].statslock = true;
            Character.updateDays(cid);
        }
        Rules.resetAddedSkills();
        Character.resetOtherSkillsList(cid,ineditmode);

        // Schema update processing
        var name;
        if (character[cid].schemaVersion === 1) {
            character[cid].schemaVersion = 2;
        } else {
            for(i=0; i<character[cid].newSkills.length; i++) {
                for(name in character[cid].newSkills[i]) {
                    if (character[cid].newSkills[i].hasOwnProperty(name)) {
                        Rules.addSkill(name,character[cid].newSkills[i][name]);
                    }
                }
            }
        }
        // Rebuild the newSkills list, (to remove duplicates and to ensure the added skills remain even when building lots of characters from the same instance
        character[cid].newSkills = [];
        for(name in Rules.skilllist["Added skills"]) {
            if (Rules.skilllist["Added skills"].hasOwnProperty(name)) {
                Character.addOtherSkill(cid,name,Rules.skilllist["Added skills"][name],ineditmode);
            }
        }

        // Continue with the schema updates
        if (character[cid].schemaVersion === 2) {
            character[cid].schemaVersion = 3;
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].birthmonth);
        }
        if (character[cid].schemaVersion === 3) {
            character[cid].schemaVersion = 4;
            character[cid].armour = Rules.armour[0];
            character[cid].shield = Rules.shields[0];
            character[cid].shieldLocations = [];
            character[cid].favouredWeapons = [];
        }
        if (character[cid].schemaVersion === 4) {
            character[cid].schemaVersion = 5;
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].birthmonth);
        }
        if (character[cid].schemaVersion === 5) {
            character[cid].schemaVersion = 6;
            character[cid].birthday = character[cid].genre.dayMonthYearAsText(character[cid].birthtime,character[cid].birthmonth);
            character[cid].birthmonth = character[cid].genre.asMonth(character[cid].birthtime);
            character[cid].birthmonthtext = character[cid].genre.dateMonthAsText(character[cid].birthmonth);
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
            character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
        }
        if (character[cid].schemaVersion === 6) {
            character[cid].schemaVersion = 7;
            character[cid].playerName = document.getElementById("c"+cid+"-playernametextarea").textContent;
        }
        if (character[cid].schemaVersion < 9) {
            character[cid].schemaVersion = 9;
            if (character[cid].genre.equals(new Genre('Linrodeth'))) {
                character[cid].birthmonth--;
                character[cid].gamemonth--;
            }
            for(i=0; i<character[cid].periods.length; i++) {
                character[cid].periods[i].startmonthtext = character[cid].genre.dateMonthAsText(character[cid].periods[i].startmonth+character[cid].birthmonth);
                character[cid].periods[i].endmonthtext = character[cid].genre.dateMonthAsText(character[cid].periods[i].endmonth+character[cid].birthmonth);
            }
            character[cid].birthmonthtext = character[cid].genre.dateMonthAsText(character[cid].birthmonth);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
        }
        if (character[cid].schemaVersion === 9) {
            character[cid].schemaVersion = 10;
            character[cid].morale = Rules.startingMorale(character[cid]); // rounding change
        }
        if (character[cid].schemaVersion === 10) {
            character[cid].schemaVersion = 11;
            character[cid].syncWithGlobalDate = (character[cid].gametime > 1200000000000);
        }
        if (character[cid].schemaVersion === 11) {
            character[cid].schemaVersion = 12;
            character[cid].statsswapable = !(character[cid].statslock);
        }
        if (character[cid].schemaVersion === 12) {
            character[cid].schemaVersion = 13;
            character[cid].gender = '';
        }
        // For Tekumel convert Long Bow to Short Bow in equipment
        if (character[cid].schemaVersion === 13) {
            character[cid].schemaVersion = 14;
            if (character[cid].genre.equals(new Genre('Tekumel'))) {
                let shortBow;
                for(let weaponIndex=0; weaponIndex < Rules.weapons.length; weaponIndex++) {
                    if (Rules.weapons[weaponIndex].name === 'Short Bow') {
                        shortBow = Rules.weapons[weaponIndex];
                        break;
                    }
                }
                for(let weaponIndex=0; weaponIndex < character[cid].favouredWeapons.length; weaponIndex++) {
                    if (character[cid].favouredWeapons[weaponIndex].name === 'Long Bow') {
                        character[cid].favouredWeapons[weaponIndex] = shortBow;
                    }
                }
            }
        }

        // Data resets - resolve skills with
        var ampregexp = new RegExp('&amp;', 'g');
        for(i=0; i<character[cid].periods.length; i++) {
            for(var j=0; j<character[cid].periods[i].interests.length; j++) {
                character[cid].periods[i].interests[j].skill = character[cid].periods[i].interests[j].skill.replace(ampregexp,'&');
            }
        }

        // Global updates
        if (character[cid].syncWithGlobalDate) {
            character[cid].gametime = options.gametime[character[cid].genre.name];
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
            character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
            character[cid].periods[character[cid].periods.length-1].endmonth = character[cid].gamemonth - character[cid].birthmonth;
            character[cid].periods[character[cid].periods.length-1].endmonthtext = character[cid].gamemonthtext;
        }
        Character.updateDays(cid); // Get the text for birthday and gameday in sync

        if (populate) {
            Character.populateAll(cid,ineditmode);
        }
    } catch(e) {
        if (e.stack) {
            alert(e + "\n\nStack Trace\n-----------------\n" + e.stack);
        } else {
            alert(e);
        }
    }
};

Character.setupButtons = function(cid) {
	// No buttons - return false to show this character is NOT, in fact,  being edited
    if (document.getElementById("c"+cid+"-birthdayminus1year") === null) return false;

    document.getElementById("c"+cid+"-birthdayminus1year").onclick = Character.birthdaydelta(cid,-365*24*60*60*1000);
    document.getElementById("c"+cid+"-birthdayminus1month").onclick = Character.birthdaydelta(cid,-30*24*60*60*1000);
    document.getElementById("c"+cid+"-birthdayminus1day").onclick = Character.birthdaydelta(cid,-24*60*60*1000);
    document.getElementById("c"+cid+"-birthdayplus1day").onclick = Character.birthdaydelta(cid,24*60*60*1000);
    document.getElementById("c"+cid+"-birthdayplus1month").onclick = Character.birthdaydelta(cid,30*24*60*60*1000);
    document.getElementById("c"+cid+"-birthdayplus1year").onclick = Character.birthdaydelta(cid,365*24*60*60*1000);

    document.getElementById("c"+cid+"-gamedayminus1year").onclick = Character.gamedaydelta(cid,-365*24*60*60*1000);
    document.getElementById("c"+cid+"-gamedayminus1month").onclick = Character.gamedaydelta(cid,-30*24*60*60*1000);
    document.getElementById("c"+cid+"-gamedayminus1day").onclick = Character.gamedaydelta(cid,-24*60*60*1000);
    document.getElementById("c"+cid+"-gamedayplus1day").onclick = Character.gamedaydelta(cid,24*60*60*1000);
    document.getElementById("c"+cid+"-gamedayplus1month").onclick = Character.gamedaydelta(cid,30*24*60*60*1000);
    document.getElementById("c"+cid+"-gamedayplus1year").onclick = Character.gamedaydelta(cid,365*24*60*60*1000);

    document.getElementById("c"+cid+"-gamedatesync").onchange = Character.returnFalse(function() {
        character[cid].syncWithGlobalDate = document.getElementById("c"+cid+"-gamedatesync").checked;
        document.getElementById("c"+cid+"-gamedayminus1year").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayminus1month").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayminus1day").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayplus1day").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayplus1month").disabled = character[cid].syncWithGlobalDate;
        document.getElementById("c"+cid+"-gamedayplus1year").disabled = character[cid].syncWithGlobalDate;
        if (character[cid].syncWithGlobalDate) {
            character[cid].gametime = options.gametime[character[cid].genre.name];
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
            character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
            character[cid].periods[character[cid].periods.length-1].endmonth = character[cid].gamemonth - character[cid].birthmonth;
            character[cid].periods[character[cid].periods.length-1].endmonthtext = character[cid].gamemonthtext;
            Character.incrementVersion(cid);
            Character.populateAll(cid,true);
        }
    });

    document.getElementById("c"+cid+"-genre").onchange = Character.returnFalse(function() {
        var sc = document.getElementById("c"+cid+"-genre");
        var oldgenre = character[cid].genre;
        character[cid].genre = new Genre(sc[sc.selectedIndex].text);
        if (!character[cid].genre.equals(oldgenre)) {
            // Check whether we need to sync with the game date or just convert the existing date
            if (character[cid].syncWithGlobalDate) {
                character[cid].gametime = options.gametime[character[cid].genre.name];
            }

            // Does this make the character too young (or even have a negative age) - if so we need to zap the birth day too
            if (character[cid].gametime - 568036800000 < character[cid].birthtime) {
                character[cid].birthtime  = character[cid].gametime - 568036800000;
                character[cid].periods[character[cid].periods.length-1].endmonth = character[cid].gamemonth - character[cid].birthmonth;
            }

            // Update the childhood language
            for(var i=0; i<character[cid].periods.length; i++) {
                character[cid].periods[i].childhoodLanguage = character[cid].genre.getDefaultLanguage();
            }

            // Trigger an update of the month names etc., by causing a zero size gamedaydelta and then a birthdaydelta
            Character.gamedaydelta(cid,0)(); // birthdayDelta returns a returnFalse function, so that needs to be invoked!
            Character.birthdaydelta(cid,0)(); // birthdayDelta returns a returnFalse function, so that needs to be invoked!

            // Reset equipment
            character[cid].armour = Rules.armour[0];
            character[cid].shield = Rules.shields[0];
            character[cid].shieldLocations = [];
            character[cid].favouredWeapons = [];

            Character.updateDays(cid);
            Character.setTextArea(cid,"djv_genre",character[cid].genre.toString());
            Character.incrementVersion(cid);
            Character.populateAll(cid,true);
        }
    });

    document.getElementById("c"+cid+"-gender").onchange = Character.returnFalse(function() {
        var sc = document.getElementById("c"+cid+"-gender");
        var oldgender = character[cid].gender;
        character[cid].gender = sc[sc.selectedIndex].text;
        if (character[cid].gender !== oldgender) {
            Character.incrementVersion(cid);
            Character.populateJSON(cid,true);
        }
    });

    document.getElementById("c"+cid+"-nametextarea").onchange = Character.returnFalse(function() {
        character[cid].name =  document.getElementById("c"+cid+"-nametextarea").value;
        Character.setTextArea(cid,"djv_nametextarea",character[cid].name);
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });

    document.getElementById("c"+cid+"-playernametextarea").onchange = Character.returnFalse(function() {
        character[cid].playerName =  document.getElementById("c"+cid+"-playernametextarea").value;
        Character.setTextArea(cid,"djv_playernametextarea",character[cid].playerName);
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });

    document.getElementById("c"+cid+"-characterdescriptiontextarea").onchange = Character.returnFalse(function() {
        character[cid].description =  document.getElementById("c"+cid+"-characterdescriptiontextarea").value;
        Character.incrementVersion(cid);
    });

    for(var i=0; i<Character.allstats.length; i++) {
        var stat = Character.allstats[i];
        document.getElementById("c"+cid+"-"+stat+"swapsel").onchange = Character.statSwap(cid,stat,document.getElementById("c"+cid+"-"+stat+"swapsel"));
    }

    document.getElementById("c"+cid+"-reroll").onclick = Character.returnFalse(function() {
        var newcharacter = Rules.newCharacter();
        for(var i=0; i<Character.allstats.length; i++) {
            character[cid][Character.allstats[i]] = newcharacter[Character.allstats[i]];
        }
        character[cid].characterVersion = 0;
        character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
        Character.updateDerivedStats(cid);
        Character.populateStats(cid);
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });

    document.getElementById("c"+cid+"-statsedit").onclick = Character.returnFalse(function(){
        var updateStat = function(stat,element) {
            return Character.returnFalse(function() {
                character[cid][stat] = Character.validateStat(stat,element.value);
                Character.updateDerivedStats(cid);
                Character.incrementVersion(cid);
            });
        };
        character[cid].statsswapable = false;
        Character.populateStats(cid);
        for(var i=0; i<Character.allstats.length; i++) {
            var oldE = document.getElementById("c"+cid+"-stats-"+Character.allstats[i]);
            var newE = document.createElement("td");
            var textarea = document.createElement("textarea");
            textarea.value = character[cid][Character.allstats[i]];
            textarea.rows = 1;
            textarea.cols = 8;
            textarea.className = "onerow";
            textarea.onchange = updateStat(Character.allstats[i],textarea);
            textarea.id = "c"+cid+"-stats-"+Character.allstats[i];
            newE.appendChild(textarea);
            oldE.parentNode.replaceChild(newE,oldE);
            Character.deleteElement(cid,Character.allstats[i]+"swap");
        }
        document.getElementById("c"+cid+"-statsedit").disabled = true;
    });

    document.getElementById("c"+cid+"-statsedit").disabled = false; // Ensure enabled even after a reload of a page with it disabled

    document.getElementById("c"+cid+"-statslock").onclick = Character.returnFalse(function() {
        character[cid].statsswapable = false;
        character[cid].statslock = true;
        for(var i=0; i<Character.allstats.length; i++) {
            var oldE = document.getElementById("c"+cid+"-stats-"+Character.allstats[i]);
            var newE = document.createElement("td");
            var text = document.createTextNode(character[cid][Character.allstats[i]].toString());
            newE.appendChild(text);
            newE.id = "c"+cid+"-stats-"+Character.allstats[i];
            oldE.parentNode.replaceChild(newE,oldE);
        }
        Character.populateStats(cid);
        Character.incrementVersion(cid);
        Character.populateJSON(cid,true);
    });

    document.getElementById("c"+cid+"-statsunlock").onclick = Character.returnFalse(function() {
        character[cid].statslock = false;
        document.getElementById("c"+cid+"-statsedit").onclick();
    });

    document.getElementById("c"+cid+"-addPeriod").onclick = Character.returnFalse(function() {
        var op = character[cid].periods[character[cid].periods.length-1];
        var np = JSON.parse(JSON.stringify(op));
        // Split the time in two
        var duration = Math.round((op.endmonth - op.startmonth + 1)/2);
        op.endmonth = op.startmonth + duration - 1;
        np.startmonth = op.endmonth + 1;
        op.endmonthtext = character[cid].genre.dateMonthAsText(op.endmonth+character[cid].birthmonth);
        np.startmonthtext = character[cid].genre.dateMonthAsText(np.startmonth+character[cid].birthmonth);
        np.title = "Period "+(character[cid].periods.length+1);
        np.description = "The next stage in life";
        character[cid].periods.push(np);
        Character.incrementVersion(cid);
        Character.populateAll(cid,true);
    });

    document.getElementById("c"+cid+"-startPeriod").onclick = Character.returnFalse(function() {
        var op = character[cid].periods[character[cid].periods.length-1];
        var np = JSON.parse(JSON.stringify(op));
        // Split the time in two
        np.startmonth = op.endmonth + 1;
        np.endmonth = op.endmonth;
        np.endmonthtext = character[cid].genre.dateMonthAsText(np.endmonth+character[cid].birthmonth);
        np.startmonthtext = character[cid].genre.dateMonthAsText(np.startmonth+character[cid].birthmonth);
        np.title = "Period "+(character[cid].periods.length+1);
        np.description = "The next stage in life";
        character[cid].periods.push(np);
        Character.incrementVersion(cid);
        Character.populateAll(cid,true);
    });

    document.getElementById("c"+cid+"-JSONtriggerfile").onclick = Character.returnFalse(function() {
        document.getElementById("c"+cid+"-JSONfile").click();
    });

    document.getElementById("c"+cid+"-JSONfile").onchange = Character.returnFalse(function() {
        var file = document.getElementById("c"+cid+"-JSONfile").files[0];
        if (file) {
            var reader = new FileReader();
            reader.onload = function() {
                if (document.getElementById("c"+cid+"-statedit")) {
                    document.getElementById("c"+cid+"-statsedit").checked = false;
                    document.getElementById("c"+cid+"-statsedit").onchange();
                }
                Character.updateWithNewJSON(cid,reader.result,true,true);
                // Copy across the form fields to the ones actually sent back to WordPress (in the hidden diff)
                Character.populateJSON(cid,true);
                Character.setTextArea(cid,"djv_nametextarea",character[cid].name);
                Character.setTextArea(cid,"djv_genre",character[cid].genre.toString());
                Character.setTextArea(cid,"djv_playernametextarea",character[cid].playerName);
            };
            reader.onerror = function() {
                alert("Error reading JSON, error code:"+reader.error);
            };
            reader.readAsText(file);
        }
    });

    return true; /* Indicate that we did setup the buttons for this character) */
};
